/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_buf.h
 * @page doca buf
 * @defgroup DOCACore Core
 * @defgroup BUF DOCA Buffer
 * @ingroup DOCACore
 * The DOCA Buffer is used for reference data. It holds the information on a memory region that belongs to
 * a DOCA memory map, and its descriptor is allocated from DOCA Buffer Inventory.
 *
 * @{
 */

#ifndef DOCA_BUF_H_
#define DOCA_BUF_H_

#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>

#include <doca_compat.h>
#include <doca_error.h>

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 * DOCA Buffer element
 ******************************************************************************/

/**
 * @brief Opaque structure representing a data buffer, that can be read by registered DOCA devices.
 */
struct doca_buf;

enum doca_buf_extension {
	DOCA_BUF_EXTENSION_NONE = 0,
};

/**
 * @brief Increase the object reference count by 1.
 *
 * @param [in] buf
 * DOCA Buf element.
 * @param [out] refcount
 * The number of references to the object before this operation took place.
 *
 * @return
 * - DOCA_ERROR_NOT_SUPPORTED
 *
 * @note This function is not supported yet.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_buf_refcount_add(struct doca_buf *buf, uint16_t *refcount);

/**
 * @brief Decrease the object reference count by 1, if 0 reached, return the element back to the inventory.
 *
 * @details When refcont 0 reached, all related resources should be released. For example if the element points into
 * some mmap its state will be adjusted accordingly.
 *
 * @param [in] buf
 * DOCA Buf element.
 * @param [out] refcount
 * The number of references to the object before this operation took place.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_buf_refcount_rm(struct doca_buf *buf, uint16_t *refcount);

/**
 * @brief Get the reference count of the object.
 *
 * @param [in] buf
 * DOCA Buf element.
 * @param [out] refcount
 * The number of references to the object.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_buf_refcount_get(struct doca_buf *buf, uint16_t *refcount);

/**
 * @brief Get the length of the payload buffer pointed by the object.
 *
 * @param [in] buf
 * DOCA Buf element.
 * @param [out] len
 * The length of the payload buffer pointed to by buf.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_buf_len_get(struct doca_buf *buf, size_t *len);

/**
 * @brief Get the payload buffer pointed by the object.
 *
 * @param [in] buf
 * DOCA Buf element.
 * @param [out] head
 * The address of the payload buffer.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_buf_head_get(struct doca_buf *buf, void **head);

#ifdef __cplusplus
}
#endif

/** @} */

#endif /* DOCA_BUF_H_ */
