/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_comm_channel.h
 * @page comm_channel
 * @defgroup DOCA_COMM_CHANNEL Comm Channel
 *
 * DOCA Communication Channel library let you set a direct communication channel between the host and the DPU.
 * The channel is run over RoCE/IB protocol and is not part of the TCP/IP stack.
 * Please follow the programmer guide for usage instructions.
 *
 * @{
 */

#ifndef DOCA_CC_H
#define DOCA_CC_H

#include <stddef.h>
#include <stdint.h>
#include <doca_error.h>
#include <doca_compat.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Service name length includes the terminating null byte ('\0')
 */
#define SERVICE_NAME_MAX 120

/**
 * @brief Flags for endpoint initialization
 */
enum doca_comm_channel_init_flags {
	DOCA_CC_INIT_FLAG_NONBLOCK	= 0x1	/**<  EP API will be non-blocking (default is blocking API calls) */
};

/**
 * @brief Flags for send/receive functions
 */
enum doca_comm_channel_msg_flags {
	DOCA_CC_MSG_FLAG_DONTWAIT	= 0x1,	/**< Enables nonblocking operations per send and/or recv calls.
						  *  if the operation would block, EAGAIN is returned */
	DOCA_CC_MSG_FLAG_MORE		= 0x2	/**< Not supported */
};

/**
 * @brief Handle for local endpoint
 */
struct doca_comm_channel_ep_t;

/**
 * @brief Handle for peer address
 */
struct doca_comm_channel_addr_t;

/**
 * @brief Configuration attributes for endpoint initialization.
 */
struct doca_comm_channel_init_attr {
	uint32_t	cookie;		/**< Cookie returned when polling the event_channel. */
	uint32_t	flags;		/**< Flags: 0 or DOCA_CC_INIT_FLAG_NONBLOCK. */
	uint32_t	maxmsgs;	/**< Max. # of messages on queue. */
	uint32_t	msgsize;	/**< Max. message size (bytes). */
};

/**
 * @brief Create local endpoint
 * The endpoint handle represents all the configuration needed for the channel to run.
 * The user needs to hold one endpoint for all actions with the CommChannel on his side
 * @param attr
 * Attributes to use when initializing the endpoint resources and QPs.
 * @param  ep
 * Output, handle to the newly created endpoint object.
 * @return
 * If the creation is successfull, ep will point to the newly created endpoint and DOCA_SUCCESS will be returned.
 * Errors:
 * DOCA_ERROR_INVALID_VALUE if no ep pointer or no attribute object was given.
 * DOCA_ERROR_NOT_PERMITTED if the given msgsize is bigger than supported max size.
 * DOCA_ERROR_NO_MEMORY if memory allocation failed during ep creation.
 * DOCA_ERROR_INITIALIZATION if initialization of ep failed.
 * DOCA_ERROR_UNKNOWN if an unknown error occurred.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_ep_create(struct doca_comm_channel_init_attr *attr, struct doca_comm_channel_ep_t **ep);

/**
 * @brief endpoint notification file descriptor for blocking with epoll() for recv ready event
 */
#ifdef __linux__
typedef int doca_event_channel_t;
#else  /**< Windows */
typedef HANDLE doca_event_channel_t;
#endif

/**
 * @brief Extract the event_channel handle for user's use
 * When the user send/receive packets with non-blocking mode, this handle can be used to get interrupt when a new
 * event happened, using epoll() or similar function.
 * The event channel is owned by the endpoint and release when calling @ref doca_comm_channel_ep_destroy().
 *
 * @param local_ep
 * handle for the endpoint created beforehand with doca_comm_channel_ep_create().
 */
__DOCA_EXPERIMENTAL
doca_event_channel_t doca_comm_channel_ep_event_channel_get(struct doca_comm_channel_ep_t *local_ep);

/**
 * @brief Service side listen on all interafces.
 *
 * The function opens new QP for each vhca_id (gvmi) it exposes to and wait for new connections.
 * After calling this function the user should call doca_comm_channel_ep_recvfrom() in order to get
 * new peers to communicate with.
 *
 * This function available only for service side use.
 *
 * @param local_ep
 * handle for the endpoint created beforehand with doca_comm_channel_ep_create().
 * @param name
 * identifies the service, as a SERVICE_NAME_MAX bytes size string.
 * @return
 * DOCA_SUCCESS on success
 * Errors:
 * DOCA_ERROR_INVALID_VALUE if no ep object or no name was given.
 * DOCA_ERROR_NOT_PERMITTED if the function was called on the client side.
 * DOCA_ERROR_NO_MEMORY if memory allocation failed.
 * DOCA_ERROR_INITIALIZATION if initialization of service failed.
 * DOCA_ERROR_CONNECTION_ABORTED if registration of service failed.
 * DOCA_ERROR_UNKNOWN if an unknown error occurred.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_ep_listen(struct doca_comm_channel_ep_t *local_ep, const char *name);

/**
 * @brief Client side Connect
 *
 * This function available only for client-side use.
 * As part of the connection process, the client send a "hello" message to the service to inform
 * him about new connection.
 *
 * If the connect function is being called before the service perform listen with the same name
 * the connection will fail.
 *
 * @param local_ep
 * handle for the endpoint created beforehand with doca_comm_channel_ep_create().
 * @param name
 * identifies the service, as a SERVICE_NAME_MAX bytes size string.
 * @param peer_addr
 * Output, handle to use for sending packets and recognize source of messages.
 * @return
 * DOCA_SUCCESS on success.
 * Errors:
 * DOCA_ERROR_INVALID_VALUE if no ep object, name or peer_address pointer given.
 * DOCA_ERROR_NOT_PERMITTED if the function was called on the service or the endpoint is already connected.
 * DOCA_ERROR_NO_MEMORY if memory allocation failed.
 * DOCA_ERROR_INITIALIZATION if initialization of ep connection failed.
 * DOCA_ERROR_CONNECTION_ABORTED if connection failed for any reason (connections rejected or failed).
 * DOCA_ERROR_UNKNOWN if an unknown error occurred.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_ep_connect(struct doca_comm_channel_ep_t *local_ep, const char *name,
					  struct doca_comm_channel_addr_t **peer_addr);

/**
 * @brief Send message to peer address.
 * The connection to the wanted peer_address need to be established before sending the message.
 *
 * @param local_ep
 * handle for the endpoint created beforehand with doca_comm_channel_ep_create().
 * @param msg
 * pointer to the message to be sent.
 * @param len
 * length in bytes of msg.
 * @param flags
 * DOCA_CC_MSG_FLAG_DONTWAIT to return on any case or 0 to block when waiting for credits to arrive.
 * @param peer_addr
 * destination address handle of the send operation.
 * @return
 * DOCA_SUCCESS on success.
 * Errors:
 * DOCA_ERROR_NOT_CONNECTED if no peer_address was supplied or no connection was found.
 * DOCA_ERROR_INVALID_VALUE if the supplied len was larger than the msgsize given at ep creation or any of the input
 *                          variables are null.
 * DOCA_ERROR_AGAIN if the command or the endpoint is set to non-blocking mode and the send queue is full.
 * 		    when returned, the user can use the endpoint's doca_event_channel_t to get indication for a new
 * 		    empty slot.
 * DOCA_ERROR_CONNECTION_RESET if the provided peer_addr experienced an error and it needs to be disconnected.
 * DOCA_ERROR_INITIALIZATION if initialization of the DCI after a send error failed
 * DOCA_ERROR_UNKNOWN if an unknown error occurred.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_ep_sendto(struct doca_comm_channel_ep_t *local_ep, const void *msg,
				size_t len, int flags,
				struct doca_comm_channel_addr_t *peer_addr);

/**
 * @brief Receive message from connected client/service.
 *
 * On service side, doca_comm_channel_ep_recvfrom() also used for accepting new connection from clients.
 *
 * @param local_ep
 * handle for the endpoint created beforehand with doca_comm_channel_ep_create().
 * @param msg
 * pointer to the buffer where the message should be stored.
 * @param len
 * input - maximum len of bytes in the msg buffer, output - len of actual received message.
 * @param flags
 * DOCA_CC_MSG_FLAG_DONTWAIT to return on any case or 0 to block when waiting on empty queue.
 * @param peer_addr
 * output, received message source address handle
 * @return
 * DOCA_SUCCESS on successful receive. If a message was received, the value pointed by len will be updated with
 * the number of bytes received.
 * Errors:
 * DOCA_ERROR_INVALID_VALUE if any of the parameters is NULL.
 * DOCA_ERROR_NOT_CONNECTED if endpoint is service and listen was not called.
 * DOCA_ERROR_AGAIN if the command or the endpoint is set to non-blocking mode and no message was received.
 * 		    when returned, the user can use the endpoint's doca_event_channel_t to get indication for a new
 * 		    arrival message.
 * DOCA_ERROR_CONNECTION_RESET if the message received is from a peer_addr that has error.
 * DOCA_ERROR_INITIALIZATION if initialization of the DCI after a send error failed
 * DOCA_ERROR_UNKNOWN if an unknown error occurred.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_ep_recvfrom(struct doca_comm_channel_ep_t *local_ep, void *msg, size_t *len,
				  int flags, struct doca_comm_channel_addr_t **peer_addr);

/**
 * @brief Disconnect the endpoint from the remote peer.
 * block until all resources related to peer address are freed
 * new connection could be created on the endpoint
 * @param local_ep
 * handle for the endpoint created beforehand with doca_comm_channel_ep_create().
 * @param peer_addr
 * peer address to be disconnect from.
 * @return
 * DOCA_SUCCESS on success.
 * Errors:
 * DOCA_ERROR_INVALID_VALUE if no ep was provided.
 * DOCA_ERROR_NOT_CONNECTED if there is no connection.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_ep_disconnect(struct doca_comm_channel_ep_t *local_ep,
					     struct doca_comm_channel_addr_t *peer_addr);

/**
 * @brief Release endpoint handle.
 *
 * Blocking until all queued messages are sent
 * The function close the event_channel and release all internal resources.
 * The doca_comm_channel_ep_disconnect() is included as part of the destroy process.
 *
 * @param local_ep
 * handle for the endpoint created beforehand with doca_comm_channel_ep_create().
 * @return
 * DOCA_SUCCESS on success.
 * Errors:
 * DOCA_ERROR_NOT_CONNECTED if ep does not exist.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_ep_destroy(struct doca_comm_channel_ep_t *local_ep);

/**
 * @brief Extract 'user_context' from peer_addr handle.
 * By default, the 'user_context' is set to 0 and can be change using @ref doca_comm_channel_peer_addr_user_data_set()
 *
 * @param peer_addr
 * Pointer to peer_addr to extract user_context from.
 * @param user_data
 * Output param, will contain the extracted data.
 *
 * @return
 * DOCA_SUCCESS on success.
 * DOCA_ERROR_INVALID_VALUE if peer_address or user_data is NULL.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_peer_addr_user_data_get(struct doca_comm_channel_addr_t *peer_addr, uint64_t *user_data);

/**
 * @brief Save 'user_context' in peer_addr handle
 *
 * Can be use by the user to identify the peer address got from doca_comm_channel_ep_recvfrom().
 * The user_context for new peers is initialized to 0.
 *
 * @param peer_addr
 * Pointer to peer_addr to set user_context to.
 * @param user_context
 * Data to set for peer_addr.
 *
 * @return
 * DOCA_SUCCESS on success.
 * DOCA_ERROR_INVALID_VALUE if peer_address is NULL.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_comm_channel_peer_addr_user_data_set(struct doca_comm_channel_addr_t *peer_addr, uint64_t user_context);

#ifdef __cplusplus
}
#endif

/** @} */

#endif /* DOCA_CC_H */
