/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * Definition of an abstract implementation to be used by doca crypto engines, Provides a separation between
 * implementation of actual crypto engine routines for the higher level functionality of the doca crypto api.
 */

/**
 * @file doca_crypto.h
 * @page doca crypto
 * @defgroup cryto engine
 * DOCA CRYPTO library. For more details please refer to the user guide on DOCA
 * devzone.
 *
 * @{
 */

#ifndef DOCA_CRYPTO_H_
#define DOCA_CRYPTO_H_

#include <doca_compat.h>
#include <doca_buf.h>
#include <doca_ctx.h>
#include <doca_error.h>

#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************************************
 * DOCA CRYPTO JOB TYPES
 *********************************************************************************************************************/

/**
 * Doca crypto action type enums, used to specify crypto job types.
 */
enum doca_crypto_action_type {
	DOCA_CRYPTO_JOB_SHA1 = DOCA_ACTION_CRYPTO_FIRST + 1,
	DOCA_CRYPTO_JOB_SHA256,
	DOCA_CRYPTO_JOB_SHA512,
	DOCA_CRYPTO_JOB_SHA1_PARTIAL,
	DOCA_CRYPTO_JOB_SHA256_PARTIAL,
	DOCA_CRYPTO_JOB_SHA512_PARTIAL,
};

/*********************************************************************************************************************
 * DOCA CRYPTO SHA JOB defines
 *********************************************************************************************************************/

/**
 * User response buffer length should be >= the following specified length, depending on the SHA function.
 * Both SHA1_PARTIAL and SHA1 need 20bytes.
 * Both SHA256_PARTIAL and SHA256 need 32bytes.
 * Both SHA512_PARTIAL and SHA512 need 64bytes.
 */
#define DOCA_SHA1_BYTE_COUNT       20			    /**< SHA1 calculation hex-format result length. */

#define DOCA_SHA256_BYTE_COUNT     32			    /**< SHA256 calculation hex-format result length. */

#define DOCA_SHA512_BYTE_COUNT     64			    /**< SHA512 calculation hex-format result length. */

/**
 * Flag enum for SHA job.
 */
enum doca_crypto_sha_job_flags {
	DOCA_CRYPTO_SHA_JOB_FLAGS_NONE = 0,		 /**< Default flag for all SHA job. */
	DOCA_CRYPTO_SHA_JOB_FLAGS_SHA_PARTIAL_FINAL	 /**< Only useful for doca_crypto_sha_partial_job. */
};

/**
 * DOCA SHA job definition.
 * -- "struct doca_crypto_sha_job" is used for one-shot SHA calculation.
 * -- Its typical usage is:
 *    -- construct a job: struct doca_crypto_sha_job job = {
 *                                   .base.type = DOCA_CRYPTO_JOB_SHA1,
 *                                   .req_buf   = user_req_buf,
 *                                   .resp_buf  = user_resp_buf,
 *                                   .flags     = DOCA_CRYPTO_SHA_JOB_FLAGS_NONE
 *                        };
 *    -- submit job:      doca_workq_submit(workq, &job.base);
 *    -- retrieve event:  doca_workq_progress_retrieve(workq, &event, DOCA_WORKQ_RETRIEVE_FLAGS_NONE);
 *
 * -- For doca_workq_submit() return code:
 *    -- DOCA_SUCCESS:
 *       -- The job is submitted successfully. It also means: this submitted source data cannot be freely
 *          manipulated until its response is received.
 *    -- DOCA_ERROR_INVALID_VALUE:
 *       -- Some of the job attribute members use illegal value.
 *          for example, response buffer length is < 20bytes for SHA1; request buffer length == 0, and the job
 *          type attribute is not supported.
 *    -- DOCA_ERROR_NO_MEMORY:
 *       -- The job resouce is exhausted for now, we need to call progress_retrieve() first to receive response
 *          and free job resource, then call job_submit() to try again to submit the same job.
 *    -- DOCA_ERROR_BAD_STATE:
 *       -- crypto_ctx is corrupted now, need reset.
 *
 * -- For doca_workq_progress_retrieve() return code:
 *    -- DOCA_SUCCESS:
 *       -- we get a response from SHA engine.
 *          user can utilise doca_job's user_data field to setup special data to correlate the returned event
 *          and the corresponding job.
 *    -- DOCA_ERROR_AGAIN:
 *       -- In order to get a response, we need to call progress_retrieve() again.
 *    -- DOCA_ERROR_IO_FAILED:
 *       -- abnormal occurs in the SHA engine hardware queue, crypto_ctx and workq need to be re-initialized.
 *    -- DOCA_ERROR_INVALID_VALUE:
 *       -- received invalid input.
 */
struct doca_crypto_sha_job {
	struct doca_job base;			 /**< Opaque structure. */
	struct doca_buf *req_buf;		 /**< User request.
						   *  SHA engine accessible buffer pointed to the user input data.
						   */
	struct doca_buf *resp_buf;		 /**< User response.
						   *  The response byte count can be decided by SHA algorithm
						   *  selection enum.
						   */
	uint64_t flags;				 /**< SHA job flags.
						   *  For the last segment of a doca_crypto_sha_partial_job, use
						   *  DOCA_CRPTO_SHA_JOB_FLAGS_SHA_PARTIAL_FINAL.
						   *  Otherwise, use DOCA_CRYPTO_SHA_JOB_FLAGS_NONE.
						   */
};

/**
 * Opaque structure representing a doca crypto sha_partial session object.
 */
struct doca_sha_partial_session;

/**
 * DOCA SHA_PARTIAL job definition.
 * -- "struct doca_crypto_sha_partial_job" is used for stateful SHA calculation.
 * -- Its typical usage for a job composed of 3 segments is:
 *    -- get a session handle:  doca_sha_partial_session *session;
 *                              doca_crypto_sha_partial_session_create(ctx, workq, &session);
 *    -- construct the 1st job: struct doca_crypto_sha_partial_job job = {
 *                                   .sha_job.base.type        = DOCA_CRYPTO_JOB_SHA1_PARTIAL,
 *                                   .sha_job.req_buf          = user_req_buf_of_1st_segment,
 *                                   .sha_job.resp_buf         = user_resp_buf,
 *                                   .sha_job.flags            = DOCA_CRYPTO_SHA_JOB_FLAGS_NONE,
 *                                   .session                  = session,
 *                              };
 *    -- submit 1st segment:    doca_workq_submit(workq, &job.sha_job.base);
 *    -- retrieve 1st event:    doca_workq_progress_retrieve(workq, &event, DOCA_WORKQ_RETRIEVE_FLAGS_NONE);
 *       The purpose of this call is to make sure the 1st_segment processing is finished before we can continue to
 *       send the next segment, because it is necessary to sequentially process all segment for generating correct
 *       SHA result. And the "user_resp_buf" at this moment contains garbage values.
 *    -- after the DOCA_SUCCESS event of the 1st segment is received, we can continue to submit 2nd segment:
 *    -- construct the 2nd job: struct doca_crypto_sha_partial_job job = {
 *                                   .sha_job.base.type        = DOCA_CRYPTO_JOB_SHA1_PARTIAL,
 *                                   .sha_job.req_buf          = user_req_buf_of_2nd_segment,
 *                                   .sha_job.resp_buf         = user_resp_buf,
 *                                   .sha_job.flags            = DOCA_CRYPTO_SHA_JOB_FLAGS_NONE,
 *                                   .session                  = session,
 *                              };
 *    -- submit 2nd segment:    doca_workq_submit(workq, &job.sha_job.base);
 *    -- retrieve 2nd event:    doca_workq_progress_retrieve(workq, &event, DOCA_WORKQ_RETRIEVE_FLAGS_NONE);
 *       The purpose of this call is also to make sure the 2nd_segment processing is finished.
 *       And the "user_resp_buf" at this moment still contains garbage values.
 *    -- after the DOCA_SUCCESS event of the 2nd segment is received, we can continue to submit 3rd/final segment:
 *    -- construct the 3rd job: struct doca_crypto_sha_partial_job job = {
 *                                   .sha_job.base.type        = DOCA_CRYPTO_JOB_SHA1_PARTIAL,
 *                                   .sha_job.req_buf          = user_req_buf_of_3rd_segment,
 *                                   .sha_job.resp_buf         = user_resp_buf,
 *                                   .sha_job.flags            = DOCA_CRYPTO_SHA_JOB_FLAGS_SHA_PARTIAL_FINAL,
 *                                   .session                  = session,
 *                              };
 *    -- submit 3rd segment:    doca_workq_submit(workq, &job.sha_job.base);
 *    -- retrieve 3rd event:    doca_workq_progress_retrieve(workq, &event, DOCA_WORKQ_RETRIEVE_FLAGS_NONE);
 *    -- After the DOCA_SUCCESS event of the 3rd segment is received, the whole job processing is done. We can get
 *       the expected SHA result from "user_resp_buf".
 *    -- release session:       doca_crypto_sha_partial_session_destroy(ctx, workq, session);
 *    -- During the whole process, please make sure to use the same "session" handle.
 *    -- And for the last segment, the "DOCA_CRYPTO_SHA_JOB_FLAGS_SHA_PARTIAL_FINAL" flag must be set.
 *
 * -- For doca_workq_submit() return code:
 *    -- DOCA_SUCCESS:
 *       -- The job is submitted successfully. It also means: this submitted source data cannot be freely
 *          manipulated until its response is received.
 *    -- DOCA_ERROR_INVALID_VALUE:
 *       -- Some of the job attribute members use illegal value.
 *          for example, response buffer length is < 20bytes for SHA1; request buffer length == 0, and the job
 *          type attribute is not supported.
 *    -- DOCA_ERROR_NO_MEMORY:
 *       -- The job resouce is exhausted for now, we need to call progress_retrieve() first to receive response
 *          and free job resource, then call job_submit() to try again to submit the same job.
 *    -- DOCA_ERROR_BAD_STATE:
 *       -- crypto_ctx is corrupted now, need reset.
 *
 * -- For doca_workq_progress_retrieve() return code:
 *    -- DOCA_SUCCESS:
 *       -- we get a response from SHA engine.
 *          user can utilise doca_job's user_data field to setup special data to correlate the returned event
 *          and the corresponding job.
 *    -- DOCA_ERROR_AGAIN:
 *       -- In order to get a response, we need to call progress_retrieve() again.
 *    -- DOCA_ERROR_IO_FAILED:
 *       -- abnormal occurs in the SHA engine hardware queue, crypto_ctx and workq need to be re-initialized.
 *    -- DOCA_ERROR_INVALID_VALUE:
 *       -- received invalid input.
 *
 * Note:
 * -- sha_partial_job session requirement:
 *    -- make sure the same doca_sha_partial_session used for all segments of a whole job.
 *    -- before 1st segment submission, call doca_crypto_sha_partial_session_create() to grab a session handle.
 *    -- from the 1st to the last segment submission, always reuse the same session handle.
 *    -- after the last segment processing, the session handle will be put back and cleanup automatically.
 *       Thus, After a successful stateful SHA, it is optional to call doca_crypto_sha_partial_session_destroy()
 *       to explicitly release a session handle.
 *    -- The doca_crypto_sha_partial_session_destroy() is provided to let user to free session handle at his will.
 *    -- If a session handle is released before the whole stateful SHA is finished, or if different handles
 *       are used for a stateful SHA, the job submission may fail due to job validity check failure; even the job
 *       submission successes, and the engine is not stalled, a wrong SHA result is expected.
 *    -- The "session" resource is limited, it is user's responsibility to make sure all allocated "session" handles
 *       are released.
 *    -- If "DOCA_CRYPTO_SHA_JOB_FLAGS_SHA_PARTIAL_FINAL" is not properly set, the engine will not be stalled,
 *       but a wrong SHA result is expected.
 *
 * -- sha_partial_job segment length requirement:
 *    -- only the last segment allows seg-byte-count != multiple-of-64 for sha1 and sha256.
 *       For example, for the above example code, the 1st and 2nd segment byte length must be multiple of 64.
 *    -- only the last segment allows seg-byte-count != multiple-of-128 for sha512.
 *    -- If the above requirement is not met, job_submission will fail.
 */
struct doca_crypto_sha_partial_job {
	struct doca_crypto_sha_job sha_job;	 /**< A basic sha_job. */
	struct doca_sha_partial_session *session;/**< An opaque structure for user.
						   *  Used to maintain state for stateful SHA calculation.
						   */
};


/*********************************************************************************************************************
 * DOCA CRYPTO Context
 *********************************************************************************************************************/

/**
 * Opaque structure representing a doca crypto instance.
 */
struct doca_crypto;

/**
 * Create a DOCA CRYPTO instance.
 *
 * @param [out] ctx
 * Instance pointer to be created, MUST NOT BE NULL.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - not enough memory for allocation.
 * - DOCA_ERROR_NOT_SUPPORTED - the required engine is not supported.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_crypto_create(struct doca_crypto **ctx);

/**
 * Destroy DOCA CRYPTO instance.
 *
 * @param [in] ctx
 * Instance to be destroyed, MUST NOT BE NULL.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_IN_USE - the required engine resource is not released yet. Please call doca_ctx_stop().
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_crypto_destroy(struct doca_crypto *ctx);

/**
 * Convert crypto instance into context for use with workQ
 *
 * @param [in] ctx
 * Crypto instance. This must remain valid until after the context is no longer required.
 *
 * @return
 * Non NULL - doca_ctx object on success.
 * Error:
 * - NULL.
 */
__DOCA_EXPERIMENTAL
struct doca_ctx *doca_crypto_as_ctx(struct doca_crypto *ctx);

/**
 * Get a session object used for sha_partial calculation.
 * For the same sha_partial job, user need to keep using the same sha_partial_session.
 * A session object can only be used for the same crypto_ctx and workq.
 * It cannot be shared between different crypto_ctx or different workq.
 *
 * @param [in] ctx
 * Crypto instance.
 * @param [in] workq
 * Workq instance.
 *
 * @param [out] session
 * Instance pointer to be created, MUST NOT BE NULL.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - the resource is exhausted for now, please wait the existing jobs to finish or
 *                          call doca_crypto_sha_partial_session_destroy() to release the allocated session objects.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_crypto_sha_partial_session_create(struct doca_crypto *ctx,
						    struct doca_workq *workq,
						    struct doca_sha_partial_session **session);

/**
 * Release the session object used for sha_partial calculation.
 * Please make sure the session to be released belonging to the specific crypto_ctx and workq.
 *
 * @param [in] ctx
 * Crypto instance.
 * @param [in] workq
 * Workq instance.
 * @param [in] session
 * A sha_partial_session object to be released.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_crypto_sha_partial_session_destroy(struct doca_crypto *ctx,
						     struct doca_workq *workq,
						     struct doca_sha_partial_session *session);

/** @} */

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DOCA_CRYPTO_H_ */
