/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_ctx.h
 * @page doca ctx
 * @defgroup CTX DOCA Context
 * @ingroup DOCACore
 * DOCA CTX is the base class of every data-path library in DOCA.
 * It is a specific library/SDK instance object providing abstract data processing functionality.
 * The library exposes events and/or jobs that manipulate data.
 *
 * @{
 */

#ifndef DOCA_CTX_H_
#define DOCA_CTX_H_

#include <stdint.h>

#include <doca_dev.h>
#include <doca_error.h>

#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************************************
 * DOCA Job
 *********************************************************************************************************************/

/**
 * @brief Power 2 single SDK/context action type range.
 */
#define DOCA_ACTION_SDK_RANGE 16

enum doca_action_type {
	DOCA_ACTION_NONE = 0,
	DOCA_ACTION_DMA_FIRST = 1 << DOCA_ACTION_SDK_RANGE,
	DOCA_ACTION_DEVEMU_FIRST = 2 << DOCA_ACTION_SDK_RANGE,
	DOCA_ACTION_DEVEMU_BLOCK_FIRST = 3 << DOCA_ACTION_SDK_RANGE,
	DOCA_ACTION_VIRTIO_FIRST = 4 << DOCA_ACTION_SDK_RANGE,
	DOCA_ACTION_VBLK_FIRST = 5 << DOCA_ACTION_SDK_RANGE,
	DOCA_ACTION_REGEX_FIRST = 6 << DOCA_ACTION_SDK_RANGE,
	DOCA_ACTION_NVME_FIRST = 7 << DOCA_ACTION_SDK_RANGE,
	DOCA_ACTION_APPSHIELD_FIRST = 8 << DOCA_ACTION_SDK_RANGE,
	DOCA_ACTION_CRYPTO_FIRST = 9 << DOCA_ACTION_SDK_RANGE,
};

union doca_data {
	void *ptr;
	uint64_t u64;
};

enum doca_job_flags {
	DOCA_JOB_FLAGS_NONE = 0,
};

/**
 * @brief Job structure describes request arguments for service provided by context.
 *
 * A context of given type may serve one or more request types defined as action type
 * (see definition of enum doca_action_type).
 *
 * DOCA Job layout
 *
 *  SDK job --> +--------------------------+
 *              |     DOCA Job (base)      |
 *              | type                     |
 *              | flags                    |
 *              | ctx                      |
 *              | user data                |
 *              |                          |
 *              +------------+-------------+ <--  job arguments
 *              |                          |      variable size
 *              | arguments                |      SDK specific
 *              |     .                    |      structure
 *              |     .                    |
 *              |     .                    |
 *              |     .                    |
 *              |     .                    |
 *              |     .                    |
 *              |                          |
 *              +------------+-------------+
 */
struct doca_job {
	int type;		   /**< Defines the type of the job. */
	int flags;		   /**< Job submission flags (see `enum doca_job_flags`). */
	struct doca_ctx *ctx;	   /**< Doca CTX targeted by the job. */
	union doca_data user_data; /**< Job identifier provided by user. Will be returned back on completion. */
};

/**
 * @brief Activity completion event.
 *
 * Event structure defines activity completion of:
 * 1. Completion event of submitted job.
 * 2. CTX received event as a result of some external activity.
 */
struct doca_event {
	int type;		    /**< The type of the event originating activity. */
	union doca_data user_data;  /**< Defines the origin of the given event.
				      *  For events originating from submitted jobs,
				      *  this will hold the same user_data provided as part of the job.
				      *  For events originating from external activity,
				      *  refer to the documentation of the specific event type. */
	union doca_data result;	    /**< Event result defined per action type arguments.
				      *  If the result is as small as 64 bit (E.g., status or similar),
				      *  it can be accessed as result.u64.
				      *  Otherwise the data is pointed to by result.ptr,
				      *  where the size is fixed for each action type.*/
};

/*********************************************************************************************************************
 * DOCA Work Queue
 *********************************************************************************************************************/

struct doca_workq;

/**
 * @brief Creates empty DOCA WorkQ object with default attributes.
 *
 * @details The returned WorkQ needs to be added to at least one DOCA CTX.
 * Then the WorkQ can be used to progress jobs and to poll events exposed by the associated CTX.
 *
 * @param [in] depth
 * The maximum number of inflight jobs.
 * @param [out] workq
 * The newly created WorkQ.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - invalid input received.
 * - DOCA_ERROR_NO_MEMORY - failed to allocate WorkQ.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_workq_create(uint32_t depth, struct doca_workq **workq);

/**
 * @brief Destroy a DOCA WorkQ.
 *
 * @details In order to destroy a WorkQ, at first needs to be removed from all DOCA CTXs using it.
 *
 * @param [in] workq
 * The WorkQ to destroy.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - invalid input received.
 * - DOCA_ERROR_IN_USE - WorkQ not removed from one of the doca_ctx.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_workq_destroy(struct doca_workq *workq);

/**
 * @brief Submit a job to a DOCA WorkQ.
 *
 * @details This method is used to submit a job to the WorkQ.
 * The WorkQ should be added to the job->ctx via doca_ctx_workq_add() before job submission.
 * Once a job has been submitted, it can be progressed using doca_workq_progress_retrieve() until the result is ready
 * and retrieved.
 *
 * @param [in] workq
 * The DOCA WorkQ used for progress and retrieval of jobs.
 * @param [in] job
 * The job to submit, the job must be compatible with the WorkQ.
 *
 * @return
 * DOCA_SUCCESS - in case the job was submitted successfully, doca_workq_progress_retrieve() can be called next.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - in case the queue is full. See WorkQ depth.
 * - DOCA_ERROR_BAD_STATE - in case job->ctx is stopped.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_workq_submit(struct doca_workq *workq, struct doca_job const *job);

enum doca_workq_retrieve_flags {
	/**< Placeholder for flags argument of doca_workq_progress_retrieve(). */
	DOCA_WORKQ_RETRIEVE_FLAGS_NONE,
};

/**
 * @brief Progress & retrieve single pending event.
 *
 * @details Polling method for progress of submitted jobs and retrieval of events.
 *
 * NOTE: for V1 retrieve supported for single event only.
 *
 * @param [in] workq
 * The WorkQ object to poll for events.
 * @param [out] ev
 * Event structure to be filled in case an event was received.
 * @param [in] flags
 * Flags for progress/retrival operations. A combination of enum doca_workq_retrieve_flags.
 *
 * @return
 * - DOCA_SUCCESS - on successful event retrieval. ev output argument is set.
 * - DOCA_ERROR_AGAIN - no event available (ev output argument not set), try again to make more progress.
 * - DOCA_ERROR_IO_FAILED - the retrieved event is a failure event. The specific error is reported per action type.
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_workq_progress_retrieve(struct doca_workq *workq, struct doca_event *ev, int flags);

/*********************************************************************************************************************
 * DOCA Work Queue properties
 *********************************************************************************************************************/

enum doca_workq_property {
	DOCA_WORKQ_PROPERTY_UNKNOWN = 0,
	DOCA_WORKQ_DEPTH, /**< Maximum number of inflight jobs allowed. */
};

/**
 * @brief Get the value of a DOCA WorkQ property.
 *
 * @param [in] workq
 * The DOCA WorkQ.
 * @param [in] property
 * The requested property to get. See enum doca_workq_property.
 * @param [out] value
 * Where to write the current property value.
 * @param [in] size
 * The size of the property in bytes.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_workq_property_get(const struct doca_workq *workq, enum doca_workq_property property, void *value,
				     uint32_t size);

/**
 * @brief Read the current maximum number of inflight jobs allowed for a DOCA workq.
 *
 * @details The maximum number of inflight jobs allowed type: uint32_t.
 *
 * @param [in] workq
 * The workq to query.
 * @param [out] depth
 * The maximum number of inflight jobs allowed for workq.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t doca_workq_depth_read(struct doca_workq *workq, uint32_t *depth)
{
	return doca_workq_property_get(workq, DOCA_WORKQ_DEPTH, (void *)depth, sizeof(*depth));
}

/**
 * @brief Set the value of a DOCA WorkQ property.
 *
 * @param [in] workq
 * The DOCA WorkQ.
 * @param [in] property
 * The requested property to set. See enum doca_workq_property.
 * @param [in] value
 * The new value of the property.
 * @param [in] size
 * The size of the property in bytes.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_workq_property_set(struct doca_workq *workq, enum doca_workq_property property, const void *value,
				     uint32_t size);

/**
 * @brief Change the maximum number of inflight jobs allowed for a DOCA WorkQ to a given value.
 *
 * @details The maximum number of inflight jobs allowed type: uint32_t.
 *
 * @param [in] workq
 * The workq to query.
 * @param [in] new_depth
 * The new maximum number of inflight jobs allowed for workq.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t doca_workq_depth_write(struct doca_workq *workq, uint32_t new_depth)
{
	return doca_workq_property_set(workq, DOCA_WORKQ_DEPTH, (void *)&new_depth, sizeof(new_depth));
}

/*********************************************************************************************************************
 * DOCA Context
 *********************************************************************************************************************/

#define DOCA_MAX_NUM_CTX 1024 /**< Maximum number of doca_ctx allowed within an application.*/

struct doca_ctx;

/**
 * @brief Add a device to a DOCA CTX.
 *
 * @param [in] ctx
 * The CTX to add the device to.
 * @param [in] dev
 * The device to add.
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - On failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_BAD_STATE - CTX is started.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_ctx_dev_add(struct doca_ctx *ctx, struct doca_dev *dev);

/**
 * @brief Remove a device from a context.
 *
 * @param [in] ctx
 * The CTX to remove the device from. Must already hold the device.
 * @param [in] dev
 * The device to remove.
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - On failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_BAD_STATE - CTX is started.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_ctx_dev_rm(struct doca_ctx *ctx, struct doca_dev *dev);

/**
 * @brief Finalizes all configurations, and starts the DOCA CTX.
 *
 * @details After starting the CTX, it can't be configured any further.
 * Use doca_ctx_stop in order to reconfigure the CTX.
 *
 * The following become possible only after start:
 * - Adding WorkQ to CTX using doca_ctx_workq_add()
 * - Removing WorkQ from CTX using doca_ctx_workq_rm()
 * - Submitting a job using doca_workq_submit()
 *
 * The following are NOT possible after start and become possible again after calling doca_ctx_stop:
 * - Adding device to CTX using doca_ctx_dev_add()
 * - Removing device from CTX using doca_ctx_dev_rm()
 *
 * @param [in] ctx
 * The DOCA context to start.
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - In case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_ctx_start(struct doca_ctx *ctx);

/**
 * @brief Stops the context allowing reconfiguration.
 *
 * @details Once a context has started, it can't be configured any further.
 * This method should be called in case the context needs to be configured after starting.
 * For more details see doca_ctx_start().
 *
 * @param [in] ctx
 * The DOCA context to stop.
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - In case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_ctx_stop(struct doca_ctx *ctx);

/**
 * @brief Add a workQ to a context.
 *
 * @details This method adds a WorkQ to a context.
 * Once a WorkQ has been added it will start accepting jobs defined by the CTX & retrieve events from the CTX.
 * The jobs can be progressed using doca_workq_progress_retrieve().
 *
 * @param [in] ctx
 * The library instance that will handle the jobs.
 * @param [in] workq
 * The WorkQ where you want to receive job completions.
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - on failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_BAD_STATE - CTX is not started.
 * - DOCA_ERROR_IN_USE - same WorkQ already added.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_ctx_workq_add(struct doca_ctx *ctx, struct doca_workq *workq);

/**
 * @brief Remove a DOCA WorkQ from a DOCA CTX.
 *
 * @details This function can only be used after CTX is started (doca_ctx_start()).
 *
 * @param [in] ctx
 * The library instance containing the WorkQ.
 * @param [in] workq
 * The WorkQ to remove.
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - on failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_BAD_STATE - CTX is not started.
 * - DOCA_ERROR_NOT_PERMITTED - WorkQ does not exist within CTX.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_ctx_workq_rm(struct doca_ctx *ctx, struct doca_workq *workq);

#ifdef __cplusplus
}
#endif

/** @} */

#endif /* DOCA_CTX_H_ */
