/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_dev.h
 * @page doca dev
 * @defgroup DEV DOCA Device
 * @ingroup DOCACore
 * The DOCA device represents an available processing unit backed by the HW or SW implementation.
 *
 * @{
 */

#ifndef DOCA_DEV_H_
#define DOCA_DEV_H_

#include <stdint.h>

#include <doca_compat.h>
#include <doca_error.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Opaque structure representing a local DOCA Device Info instance.
 * This structure is used to query information about the local device.
 */
struct doca_devinfo;
/**
 * @brief Opaque structure representing a remote DOCA Device Info instance.
 * This structure is used to query information about the remote device.
 */
struct doca_devinfo_remote;

/**
 * @brief Opaque structure representing a local DOCA Device instance.
 * This structure is used by libraries for accessing the underlying local device.
 */
struct doca_dev;
/**
 * @brief Opaque structure representing a remote DOCA Device instance.
 * This structure is used by libraries for accessing the underlying remote device.
 */
struct doca_dev_remote;

struct doca_pci_bdf {
	union {
		uint16_t raw;
		struct {
			uint16_t bus : 8;
			uint16_t device : 5;
			uint16_t function : 3;
		};
	};
};

/**
 * Remote device filter by flavor
 *
 * Multiple options possible but some are mutually exclusive.
 *
 */
enum doca_dev_remote_filter {
	DOCA_DEV_REMOTE_FILTER_ALL		= 0,
	DOCA_DEV_REMOTE_FILTER_NET		= 1 << 1,
};

/**
 * @brief Creates list of all available local devices.
 *
 * Lists information about available devices, to start using the device you first have to call doca_dev_open(),
 * while passing an element of this list. List elements become invalid once it has been destroyed.
 *
 * @param [out] dev_list
 * Pointer to array of pointers. Output can then be accessed as follows (*dev_list)[idx].
 * @param [out] nb_devs
 * Number of available local devices.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - failed to allocate enough space.
 * - DOCA_ERROR_NOT_FOUND - failed to get RDMA devices list
 * @note Returned list must be destroyed using doca_devinfo_list_destroy()
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_devinfo_list_create(struct doca_devinfo ***dev_list, uint32_t *nb_devs);

/**
 * @brief Destroy list of local device info structures.
 *
 * Destroys the list of device information, once the list has been destroyed, all elements become invalid.
 *
 * @param [in] dev_list
 * List to be destroyed.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_IN_USE - at least one device in the list is in a corrupted state.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_devinfo_list_destroy(struct doca_devinfo **dev_list);

/**
 * @brief Create list of available remote devices accessible by dev.
 *
 * Returns all representors managed by the provided device.
 * The provided device must be a local device.
 * The representor may represent a network function attached to the host,
 * or it can represent an emulated function attached to the host.
 *
 * @param [in] dev
 * Local device with access to representors.
 * @param [in] filter
 * Bitmap filter of representor types. See enum doca_dev_remote_filter for more details.
 * @param [out] dev_list_remote
 * Pointer to array of pointers. Output can then be accessed as follows (*dev_list_remote)[idx].
 * @param [out] nb_devs_remote
 * Number of available remote devices.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - failed to allocate memory for list.
 * - DOCA_ERROR_NOT_SUPPORTED - local device does not expose remote devices.
 * @note Returned list must be destroyed using doca_devinfo_remote_list_destroy()
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_devinfo_remote_list_create(struct doca_dev *dev, int filter,
					     struct doca_devinfo_remote ***dev_list_remote,
					     uint32_t *nb_devs_remote);

/**
 * @brief Destroy list of remote device info structures.
 *
 * Destroy list of remote device information, once the list has been destroyed,
 * all elements of the list are considered invalid.
 *
 * @param [in] dev_list_remote
 * List to be destroyed.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_IN_USE - the doca_dev that created the list is in a corrupted state.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_devinfo_remote_list_destroy(struct doca_devinfo_remote **dev_list_remote);

/**
 * @brief Initialize local device for use.
 *
 * @param [in] devinfo
 * The devinfo structure of the requested device.
 * @param [out] dev
 * Initialized local doca device instance on success. Valid on success only.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - failed to allocate protection domain for device.
 * - DOCA_ERROR_NOT_CONNECTED - failed to open device.
 * @note In case the same device was previously opened, then the same doca_dev instance is returned.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_dev_open(struct doca_devinfo *devinfo, struct doca_dev **dev);

/**
 * @brief Destroy allocated local device instance.
 *
 * @param [in] dev
 * The local doca device instance.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * - DOCA_ERROR_IN_USE - failed to deallocate device resources.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_dev_close(struct doca_dev *dev);

/**
 * @brief Initialize remote device for use.
 *
 * @param [in] devinfo
 * The devinfo structure of the requested device.
 * @param [out] dev_remote
 * Initialized remote doca device instance on success. Valid on success only.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - failed to allocate memory for device.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_dev_remote_open(struct doca_devinfo_remote *devinfo, struct doca_dev_remote **dev_remote);

/**
 * @brief Destroy allocated remote device instance.
 *
 * @param [in] dev
 * The remote doca device instance.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_IN_USE - failed to deallocate device resources.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_dev_remote_close(struct doca_dev_remote *dev);

/**
 * @brief Get local device info from device.
 * This should be useful when wanting to query information about device after opening it,
 * and destroying the devinfo list.
 *
 * @param [in] dev
 * The doca device instance.
 *
 * @return
 * The matching doca_devinfo instance in case of success, NULL in case dev is invalid.
 */
__DOCA_EXPERIMENTAL
struct doca_devinfo *doca_dev_as_devinfo(struct doca_dev *dev);

/**
 * @brief Get remote device info from device.
 * This should be useful when wanting to query information about device after opening it,
 * and destroying the devinfo list.
 *
 * @param [in] dev_remote
 * The remote doca device instance.
 *
 * @return
 * The matching doca_devinfo_remote instance in case of success, NULL in case dev_remote is invalid.
 */
__DOCA_EXPERIMENTAL
struct doca_devinfo_remote *doca_dev_remote_as_devinfo(struct doca_dev_remote *dev_remote);

/*********************************************************************************************************************
 * DOCA Local Device Info Properties
 *********************************************************************************************************************/

#define DOCA_DEVINFO_VUID_SIZE 128
#define DOCA_DEVINFO_IPV4_ADDR_SIZE 4
#define DOCA_DEVINFO_IPV6_ADDR_SIZE 16
#define DOCA_DEVINFO_IFACE_NAME_SIZE 256
#define DOCA_DEVINFO_IBDEV_NAME_SIZE 64

enum doca_devinfo_property {
	DOCA_DEVINFO_PROPERTY_UNKNOWN = 0,
	DOCA_DEVINFO_PROPERTY_VUID,            /**< Vendor Unique ID. Used as stable ID of a VF/PF. */
	DOCA_DEVINFO_PROPERTY_PCI_ADDR,        /**< BDF of the device - same address in lspci. */
	DOCA_DEVINFO_PROPERTY_IPV4_ADDR,       /**< IPv4 address of the device. */
	DOCA_DEVINFO_PROPERTY_IPV6_ADDR,       /**< IPv6 address of the device. */
	DOCA_DEVINFO_PROPERTY_IFACE_NAME,      /**< Name of the ethernet interface - same name in ifconfig. */
	DOCA_DEVINFO_PROPERTY_IBDEV_NAME,      /**< Name of the IB device. */
};

/**
 * @brief Get the value of a DOCA devinfo property.
 *
 * @param [in] devinfo
 * The device to query.
 * @param [in] property
 * The requested property to get. See enum doca_devinfo_property.
 * @param [out] value
 * Where to write the current property value.
 * @param [in] size
 * The size of the property in bytes.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input, or size does not match the property size.
 *                              See enum doca_devinfo_property.
 * - DOCA_ERROR_NOT_SUPPORTED - provided devinfo does not support this property.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_devinfo_property_get(const struct doca_devinfo *devinfo, enum doca_devinfo_property property,
				       void *value, uint32_t size);

/**
 * @brief Read the Vendor Unique ID of a DOCA devinfo.
 *
 * @details The Vendor Unique ID is used as stable ID of a VF/PF.
 * The Vendor Unique ID type: char[128].
 *
 * @param [in] devinfo
 * The device to query.
 * @param [out] vuid
 * The Vendor Unique ID of devinfo.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t doca_devinfo_vuid_read(const struct doca_devinfo *devinfo, char vuid[DOCA_DEVINFO_VUID_SIZE])
{
	return doca_devinfo_property_get(devinfo, DOCA_DEVINFO_PROPERTY_VUID, (void *)vuid, DOCA_DEVINFO_VUID_SIZE);
}

/**
 * @brief Read the PCI address of a DOCA devinfo.
 *
 * @details The BDF of the device - same as the address in lspci.
 * The PCI address type: struct doca_pci_bdf.
 *
 * @param [in] devinfo
 * The device to query.
 * @param [out] pci_addr
 * The PCI address of devinfo.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NOT_SUPPORTED - provided devinfo does not support this property.
 */
static inline doca_error_t doca_devinfo_pci_addr_read(const struct doca_devinfo *devinfo, struct doca_pci_bdf *pci_addr)
{
	return doca_devinfo_property_get(devinfo, DOCA_DEVINFO_PROPERTY_PCI_ADDR, (void *)pci_addr,
					 (uint32_t)sizeof(*pci_addr));
}

/**
 * @brief Read the IPv4 address of a DOCA devinfo.
 *
 * @details The IPv4 address type: uint8_t[4].
 *
 * @param [in] devinfo
 * The device to query.
 * @param [out] ipv4_addr
 * The IPv4 address of devinfo.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t doca_devinfo_ipv4_addr_read(const struct doca_devinfo *devinfo,
						       uint8_t ipv4_addr[DOCA_DEVINFO_IPV4_ADDR_SIZE])
{
	return doca_devinfo_property_get(devinfo, DOCA_DEVINFO_PROPERTY_IPV4_ADDR, (void *)ipv4_addr,
					 DOCA_DEVINFO_IPV4_ADDR_SIZE);
}

/**
 * @brief Read the IPv6 address of a DOCA devinfo.
 *
 * @details The IPv6 address type: uint8_t[16].
 *
 * @param [in] devinfo
 * The device to query.
 * @param [out] ipv6_addr
 * The IPv6 address of devinfo.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t doca_devinfo_ipv6_addr_read(const struct doca_devinfo *devinfo,
						       uint8_t ipv6_addr[DOCA_DEVINFO_IPV6_ADDR_SIZE])
{
	return doca_devinfo_property_get(devinfo, DOCA_DEVINFO_PROPERTY_IPV6_ADDR, (void *)ipv6_addr,
					 DOCA_DEVINFO_IPV6_ADDR_SIZE);
}

/**
 * @brief Read the name of the ethernet interface of a DOCA devinfo.
 *
 * @details The name of the ethernet interface is the same as it's name in ifconfig.
 * The name of the ethernet interface type: char[256].
 *
 * @param [in] devinfo
 * The device to query.
 * @param [out] iface_name
 * The name of the ethernet interface of devinfo.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t doca_devinfo_iface_name_read(const struct doca_devinfo *devinfo,
							char iface_name[DOCA_DEVINFO_IFACE_NAME_SIZE])
{
	return doca_devinfo_property_get(devinfo, DOCA_DEVINFO_PROPERTY_IFACE_NAME, (void *)iface_name,
					 DOCA_DEVINFO_IFACE_NAME_SIZE);
}

/**
 * @brief Read the name of the IB device represented by a DOCA devinfo.
 *
 * @details The name of the IB device type: char[64].
 *
 * @param [in] devinfo
 * The device to query.
 * @param [out] ibdev_name
 * The name of the IB device represented by devinfo.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t doca_devinfo_ibdev_name_read(const struct doca_devinfo *devinfo,
							char ibdev_name[DOCA_DEVINFO_IBDEV_NAME_SIZE])
{
	return doca_devinfo_property_get(devinfo, DOCA_DEVINFO_PROPERTY_IBDEV_NAME, (void *)ibdev_name,
					 DOCA_DEVINFO_IBDEV_NAME_SIZE);
}

/*********************************************************************************************************************
 * DOCA Remote Device Info Properties
 *********************************************************************************************************************/

#define DOCA_DEVINFO_REMOTE_VUID_SIZE 128

enum doca_devinfo_remote_property {
	DOCA_DEVINFO_REMOTE_PROPERTY_UNKNOWN = 0,
	DOCA_DEVINFO_REMOTE_PROPERTY_VUID,        /**< Vendor Unique ID. Used as stable ID of a VF/PF. */
};

/**
 * @brief Get the value of a DOCA remote devinfo property.
 *
 * @param [in] devinfo_remote
 * The device to query.
 * @param [in] property
 * The requested property to get. See enum doca_devinfo_remote_property.
 * @param [out] value
 * Where to write the current property value.
 * @param [in] size
 * The size of the property in bytes.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input, or size does not match the property size.
 *                              See enum doca_devinfo_remote_property.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_devinfo_remote_property_get(const struct doca_devinfo_remote *devinfo_remote,
					      enum doca_devinfo_remote_property property, void *value,
					      uint32_t size);

/**
 * @brief Read the Vendor Unique ID of a remote DOCA devinfo.
 *
 * @details The Vendor Unique ID is used as stable ID of a VF/PF.
 * The Vendor Unique ID type: char[128].
 *
 * @param [in] devinfo_remote
 * The remote device to query.
 * @param [out] remote_vuid
 * The Vendor Unique ID of devinfo_remote.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 */
static inline doca_error_t doca_devinfo_remote_vuid_read(const struct doca_devinfo_remote *devinfo_remote,
							 char remote_vuid[DOCA_DEVINFO_REMOTE_VUID_SIZE])
{
	return doca_devinfo_remote_property_get(devinfo_remote, DOCA_DEVINFO_REMOTE_PROPERTY_VUID,
						(void *)remote_vuid, DOCA_DEVINFO_REMOTE_VUID_SIZE);
}

#ifdef __cplusplus
}
#endif

/** @} */

#endif /* DOCA_DEV_H_ */
