/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_regex.h
 * @page doca regex
 * @defgroup RegEx RegEx engine
 * DOCA RegEx library. For more details please refer to the user guide on DOCA
 * devzone.
 *
 * @{
 */

#ifndef DOCA_REGEX_H_
#define DOCA_REGEX_H_

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include <doca_buf.h>
#include <doca_compat.h>
#include <doca_dev.h>
#include <doca_regex_mempool.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Opaque structure representing a DOCA RegEx instance.
 */
struct doca_regex;

/**
 * Data required to dispatch a job to a RegEx engine.
 */
struct doca_regex_job_request {
	/** ID of the job, useful to correlate the response. */
	uint64_t id;

	/**
	 * IDs which can be used to select which group of rules are used to process this job. Set each value to a non
	 * zero value to enable this feature or 0 to ignore it.
	 */
	uint16_t rule_group_ids[4];

	/** Data for the job */
	struct doca_buf const *buffer;
};

/**
 * Description of a RegEx match
 */
struct doca_regex_match {
	/** Allows matches to be linked together for easy management and iteration */
	struct doca_regex_match *next;

	/** Index relative to the start of the job / stream where the match begins */
	uint32_t match_start;

	/** ID of rule used to generate this match. */
	uint32_t rule_id;

	/** Length of matched value. */
	uint32_t length;
};

/**
 * Response status flags
 */
enum doca_regex_status_flag {
	DOCA_REGEX_STATUS_SEARCH_FAILED = 1,
};

/**
 * Result of a RegEx search
 */
struct doca_regex_job_response {
	/** ID of the enqueued job. */
	uint64_t id;

	/** Response flags. A bit masked field for zero or more status flags. See doca_regex_status_flag */
	uint64_t status_flags;

	/** Total number of detected matches. */
	uint32_t detected_matches;

	/** Total number of returned matches. */
	uint32_t num_matches;

	/** Returned matches. Contains num_matches elements as a linked list */
	struct doca_regex_match *matches;

	/** Memory pool owning the matches */
	struct doca_regex_mempool *matches_mempool;
};

/** The collection of properties that can be read or written, see individual items from more details on each */
enum doca_regex_property {
	/**
	 * Access:              WRITE
	 * Value type:          array of bytes
	 * Default value:       N/A
	 *
	 * Specify the binary (compiled) rules data to be used by the hardware regex device. Argument(value)
	 * passed should be a pointer to the first byte of an array of bytes which holds Argument(size) bytes of data.
	 */
	DOCA_REGEX_PROPERTY_HARDWARE_BINARY_RULES,

	/**
	 * Access:              WRITE
	 * Value type:          array of bytes
	 * Default value:       N/A
	 *
	 * Specify the binary (compiled) rules data to be used by the software regex device. Argument(value)
	 * passed should be a pointer to the first byte of an array of bytes which holds Argument(size) bytes of data.
	 */
	DOCA_REGEX_PROPERTY_SOFTWARE_BINARY_RULES,

	/**
	 * Access:              READ|WRITE
	 * Value type:          bool
	 * Default value:       false
	 *
	 * Enable this feature so that when a job fails to execute on the primary regex device it will be automatically
	 * re-executed on the secondary device. Jobs that fail to execute on the Secondary device will never be retried
	 * on the primary device. Set this value to true to enable the feature.
	 */
	DOCA_REGEX_PROPERTY_FAILED_JOB_FALLBACK,

	/**
	 * Access:              READ|WRITE
	 * Value type:          uint16_t
	 * Default value:       0
	 *
	 * Set the size of overlap to use when a job exceeds a devices maximum search size.
	 *
	 * When a submitted job is larger than the receiving device can support it must be fragmented. This can cause
	 * issues if a match exists but is split across two fragments. To remedy this an overlap size can be set so that
	 * these matches may be detected. The overlap defined by this function specifies how many bytes of the previous
	 * search fragment will be resent as part of the next search fragment. So for example if a 100 byte job is
	 * submitted and a device supported a 32 byte maximum job length then the jobs sent would look as follows:
	 *
	 * Overlap size   First job   Second Job   Third Job   Fourth job   Fifth Job   Sixth Job
	 * 0              [0-31]      [32-63]      [64-95]     [96-99]      ---         ---
	 * 8              [0-31]      [24-55]      [42-79]     [72-99]      ---         ---
	 * 16             [0-31]      [16-47]      [32-63]     [48-79]      [64-95]     [80-99]
	 *
	 * This allows the user to select an overlap value which provides enough overlap to detect any match they must
	 * find for the lowest cost.
	 */
	DOCA_REGEX_PROPERTY_HUGE_JOB_EMULATION_OVERLAP,

	/**
	 * Access:              READ|WRITE
	 * Value type:          uint16_t
	 * Default value:       0
	 *
	 * Define a threshold for "small jobs". For scenarios where small jobs cause poor performance using the primary
	 * regex device these can instead be redirected to the secondary device. Set this to a value > 0 to enable the
	 * feature. Set this value to 0 to disable the feature. Defaults to 0 (disabled)
	 */
	DOCA_REGEX_PROPERTY_SMALL_JOB_OFFLOAD_THRESHOLD,

	/**
	 * Access:              WRITE
	 * Value type:          uint32_t
	 * Default value:       1000
	 *
	 * Each work queue attached to the regex engine gets a pool allocator for matches. Set this value to set the
	 * maximum number of matches that can be stored for a given workq.
	 */
	DOCA_REGEX_PROPERTY_MATCHES_MEMORY_POOL_SIZE,
};

/**
 * Possible RegEx device capabilities.
 */
enum doca_regex_devinfo_caps {
	/** Device has absolutely no regex support */
	DOCA_REGEX_CAP_NONE = 0,

	/** Device is capable of RegEx operations */
	DOCA_REGEX_CAP_REGEX_SUPPORTED = 1 << 0,

	/** Hardware off-load is available */
	DOCA_REGEX_CAP_HARDWARE_OFFLOAD = 1 << 1,
};

/**
 * Check if given device is RegEx capable.
 *
 * @param [in] devinfo
 * The DOCA device information
 *
 * @param [out] caps
 * RegEx capabilities available through this device. see enum doca_regex_devinfo_caps.
 *
 * @return
 * DOCA_SUCCESS - in case at least one capability is supported.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_UNKNOWN - failed to query device capability. Maybe old FW?
 * - DOCA_ERROR_NOT_SUPPORTED - provided devinfo does not support any RegEx capability.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_regex_devinfo_caps_get(struct doca_devinfo const *devinfo, uint32_t *caps);

/**
 * Create a DOCA RegEx instance.
 *
 * @return
 * Non NULL doca_regex object on success, NULL otherwise.
 */
__DOCA_EXPERIMENTAL
struct doca_regex *doca_regex_create(void);

/**
 * Destroy DOCA RegEx instance.
 *
 * @param [in] regex
 * Instance to be destroyed, MUST NOT BE NULL.
 *
 * @note
 * The application must call doca_regex_stop() before destroying a running instance.
 */
__DOCA_EXPERIMENTAL
void doca_regex_destroy(struct doca_regex *regex);

/**
 * Make the RegEx Engine ready for processing jobs.
 *
 * @details After calling this functions the engine will not accept any further configuration but will start accepting
 * jobs. If further configuration is required the engine should be stopped first to allow this configuration to be
 * performed.
 *
 * @param [in] regex
 * The DOCA RegEx instance to start, MUST NOT be NULL..
 *
 * @return
 * 0 on success or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_start(struct doca_regex *regex);

/**
 * Stop the RegEx engine.
 *
 * @details Disable processing of jobs and return the engine to an idle state. The engine will then be ready to accept
 * new configuration or be ready to be destroyed. Any in flight jobs when this function is called will be terminated and
 * will not return any results.
 *
 * @param [in] regex
 * The DOCA RegEx instance to stop, MUST NOT be NULL.
 *
 * @return
 * 0 on success or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_stop(struct doca_regex *regex);

/**
 * Add a device to doca_regex
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] dev
 * Device to add to doca_regex. must not be NULL.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_dev_add(struct doca_regex *regex, struct doca_dev *dev);

/**
 * Remove a device from doca_regex
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] dev
 * Device to remove from doca_regex. must not be NULL.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_dev_rm(struct doca_regex *regex, struct doca_dev *dev);

/**
 * Specify how many qp's are available for use
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] num_qps
 * Number of QP's that can be used. Must be at least 1.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_regex instance is not running. see doca_regex_start,
 * doca_regex_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
int doca_regex_num_qps_set(struct doca_regex *regex, uint16_t num_qps);

/**
 * Set a property for the RegEx engine. All values are deeply copied by the RegEx Instance so the caller is free to
 * dispose of the input value at a time that is convenient for them after the function completes.
 *
 * @note
 * Properties can only be set while the RegEx instance is not running.
 *
 * @note
 * Properties are only validated in isolation as they are set. A full combination validation and integrity check of the
 * regex state is done during doca_ctx_start. As such success to set a property does not ensure that the regex instance
 * is correctly configured.
 *
 * @param [in] regex
 * The regex instance to configure
 *
 * @param [in] property
 * Enumerated property id.
 *
 * @param [in] value
 * Pointer to value. The type and size of data pointed to by this function is specified with each doca_regex_property.
 *
 * @param [in] size
 * Number of bytes of data pointed to by value.
 *
 * @return
 * DOCA_SUCCESS - RegEx instance was created
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_LOCK - Unable to gain exclusive control of RegEx instance.
 * - DOCA_ERROR_IN_USE - RegEx instance is currently started.
 * - DOCA_ERROR_NO_MEMORY - Unable to allocate memory to store a copy of property value.
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a single thread.
 */
__DOCA_EXPERIMENTAL
doca_error_t doca_regex_property_set(struct doca_regex *regex, enum doca_regex_property property, uint8_t const *value,
				     uint32_t size);

/**
 * Helper inline wrapper to make setting DOCA_REGEX_PROPERTY_HUGE_JOB_EMULATION_OVERLAP easier and less error prone
 */
static inline doca_error_t
doca_regex_property_huge_job_emulation_overlap_set(struct doca_regex *regex, uint16_t nb_overlap_bytes)
{
	return doca_regex_property_set(regex, DOCA_REGEX_PROPERTY_HUGE_JOB_EMULATION_OVERLAP,
				       (uint8_t const *)&nb_overlap_bytes, sizeof(nb_overlap_bytes));
}

/**
 * Helper inline wrapper to make setting DOCA_REGEX_PROPERTY_SMALL_JOB_OFFLOAD_THRESHOLD easier and less error prone
 */
static inline doca_error_t
doca_regex_property_small_job_offload_threshold_set(struct doca_regex *regex, uint16_t threshold)
{
	return doca_regex_property_set(regex, DOCA_REGEX_PROPERTY_SMALL_JOB_OFFLOAD_THRESHOLD,
				       (uint8_t const *)&threshold, sizeof(threshold));
}

/**
 * Helper inline wrapper to make setting DOCA_REGEX_PROPERTY_FAILED_JOB_FALLBACK easier and less error prone
 */
static inline doca_error_t
doca_regex_property_failed_job_fallback_set(struct doca_regex *regex, bool enabled)
{
	return doca_regex_property_set(regex, DOCA_REGEX_PROPERTY_FAILED_JOB_FALLBACK, (uint8_t const *)&enabled,
				       sizeof(enabled));
}

/**
 * Helper inline wrapper to make setting DOCA_REGEX_PROPERTY_HARDWARE_BINARY_RULES easier and less error prone
 */
static inline doca_error_t
doca_regex_property_hardware_binary_rules_set(struct doca_regex *regex, uint8_t const *rules_buffer,
					      size_t rules_buffer_size)
{
	return doca_regex_property_set(regex, DOCA_REGEX_PROPERTY_HARDWARE_BINARY_RULES, rules_buffer,
				       rules_buffer_size);
}

/**
 * Helper inline wrapper to make setting DOCA_REGEX_PROPERTY_SOFTWARE_BINARY_RULES easier and less error prone
 */
static inline doca_error_t
doca_regex_property_software_binary_rules_set(struct doca_regex *regex, uint8_t const *rules_buffer,
					      size_t rules_buffer_size)
{
	return doca_regex_property_set(regex, DOCA_REGEX_PROPERTY_SOFTWARE_BINARY_RULES, rules_buffer,
				       rules_buffer_size);
}

/**
 * Helper inline wrapper to make setting DOCA_REGEX_PROPERTY_MATCHES_MEMORY_POOL_SIZE easier and less error prone
 */
static inline doca_error_t
doca_regex_property_matches_memory_pool_size_set(struct doca_regex *regex, uint32_t num_mem_pool_elements)
{
	return doca_regex_property_set(regex, DOCA_REGEX_PROPERTY_MATCHES_MEMORY_POOL_SIZE,
				       (uint8_t *)&num_mem_pool_elements, sizeof(uint32_t));
}

/**
 * Enqueue a job.
 *
 * @note
 * This function is thread safe when each queue is written to by only one thread. If there is a chance that more than
 * one thread could write to the same queue then external synchronisation must be provided by the application to prevent
 * that circumstance.
 *
 * @param [in] regex
 * The RegEx engine instance to use.
 *
 * @param [in] qid
 * ID of queue to enqueue the job into.
 *
 * @param [in] job
 * The job to enqueue. Caller retains ownership of the job
 *
 * @param [in] allow_aggregation
 * When set the RegEx engine may choose to not begin processing this job immediately in an attempt to maximise overall
 * efficiency and throughput. When not set the RegEx engine must begin processing immediately, potentially reducing
 * latency. This allows an application to favour either throughput or latency. If in doubt it is recommended to favour
 * throughput.
 *
 * @return
 * [0..1] Number of jobs enqueued or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_enqueue(struct doca_regex *regex, uint16_t qid, struct doca_regex_job_request const *job,
		       bool allow_aggregation);

/**
 * Dequeue responses from the RegEx engine.
 *
 * @note
 * This function is thread safe when each queue is read by only one thread. If there is a chance that more than one
 * thread could read from to the same queue then external synchronisation must be provided by the application to prevent
 * that circumstance.
 *
 * @param [in] regex
 * The RegEx engine to dequeue data from.
 *
 * @param [in] qid
 * ID of queue to read results from.
 *
 * @param [in,out] responses
 * A pointer to an array of responses. The RegEx engine will place each dequeued result into this array until either the
 * maximum number of responses are returned or all available responses have been returned. Caller assumes ownership of
 * each response returned this way. The engine takes no care to clear any un-populated element of this array, it is the
 * caller responsibility to ensure they only read as many responses as was indicated in the functions return value.
 *
 * @param [in] max_results
 * Maximum number of results to return. the responses array MUST have capacity for at least this many elements.
 *
 * @return
 * [0..max_results] The number of responses dequeued or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_regex_dequeue(struct doca_regex *regex, uint16_t qid, struct doca_regex_job_response *responses,
		       uint8_t max_results);

#ifdef __cplusplus
} /* extern "C" */
#endif

/** @} */

#endif /* DOCA_REGEX_H_ */
