/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file doca_sha.h
 * @page doca sha
 * @defgroup SHA SHA engine
 * DOCA SHA library. For more details please refer to the user guide on DOCA
 * devzone.
 *
 * @{
 */

#ifndef DOCA_SHA_H_
#define DOCA_SHA_H_

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include <doca_compat.h>
#include <doca_sha_buffer.h>
#include <doca_sha_device.h>

#ifdef __cplusplus
extern "C" {
#endif

#define SHA1_BYTE_COUNT       20		  /**< SHA1 calculation hex-format result length. */

#define SHA256_BYTE_COUNT     32		  /**< SHA256 calculation hex-format result length. */

#define SHA512_BYTE_COUNT     64		  /**< SHA512 calculation hex-format result length. */

#define SHA_OUTPUT_BYTE_COUNT (SHA512_BYTE_COUNT) /**< Indicate sha-engine raw/temp result length. */

#define MAX_BURST_COUNT       16		  /**< Maximum allowed job burst count in a single enqueue. */

#define MAX_RESPS_COUNT       256		  /**< Maximum allowed responses count in a single dequeue. */

/** SHA algorithm selection */
enum SHA_FUNC_TYPE {
	SHA_FUNC_TYPE_SHA1 = 0,
	SHA_FUNC_TYPE_SHA256,
	SHA_FUNC_TYPE_SHA512
};


/**
 * Opaque structure representing a DOCA SHA instance.
 */
struct doca_sha;

/**
 * Data required to dispatch a sha request
 * a sha request may contains one single job or a batch of jobs to a SHA engine.
 * one job is for one sha signature.
 */
struct doca_sha_job_request {
	uint64_t id;			      /**< Customer ID of the job, useful to correlate the response. */
	enum SHA_FUNC_TYPE sha_func;	      /**< SHA algorithm selection enum. */
	struct doca_sha_buffer const *buffer; /**< Data for the jobs. */
};

/**
 * Result of SHA signatures for a single sha request
 */
struct doca_sha_job_response {
	uint64_t id;			    /**< Customer ID from the enqueued job. */
	uint32_t len;			    /**< Length of sha engine response. */
	uint8_t buf[SHA_OUTPUT_BYTE_COUNT]; /**< Data of the response. */
};

/**
 * Create a DOCA SHA instance.
 *
 * @return
 * Non NULL doca_sha object on success, NULL otherwise.
 */
__DOCA_EXPERIMENTAL
struct doca_sha *doca_sha_create(void);

/**
 * Destroy DOCA SHA instance.
 *
 * @param [in] sha
 * Instance to be destroyed, MUST NOT BE NULL.
 *
 * @note
 * The application must call doca_sha_stop() before destroying a running instance.
 */
__DOCA_EXPERIMENTAL
void doca_sha_destroy(struct doca_sha *sha);

/**
 * Make the SHA Engine ready for processing jobs.
 *
 * @details After calling this functions the engine will not accept any further configuration but
 * will start accepting jobs. If further configuration is required the engine should be stopped
 * first to allow this configuration to be performed.
 *
 * @param [in] sha
 * The DOCA SHA instance to start, MUST NOT be NULL..
 *
 * @return
 * 0 on success or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_sha_start(struct doca_sha *sha);

/**
 * Stop the SHA engine.
 *
 * @details Disable processing of jobs and return the engine to an idle state. The engine will then
 * be ready to accept new configuration or be ready to be destroyed. Any in flight jobs when this
 * function is called will be terminated and will not return any results.
 *
 * @param [in] sha
 * The DOCA SHA instance to stop, MUST NOT be NULL.
 *
 * @return
 * 0 on success or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_sha_stop(struct doca_sha *sha);

/**
 * Set the device to be used for hardware accelerated searches
 *
 * @param [in] sha
 * The SHA engine instance to use.
 *
 * @param [in] device
 * Provide a non NULL value to register a device to use for hardware accelerated searches. The
 * application retains ownership of the device object. The Application shall ensure the lifetime of
 * the device is maintained until the device is either de-registered or the doca_sha instance is
 * destroyed. Pass a NULL value to this function to de-register a registered device.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_sha instance is not running. see
 * doca_sha_start, doca_sha_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a
 * single thread.
 */
__DOCA_EXPERIMENTAL
int doca_sha_device_set(struct doca_sha *sha, struct doca_sha_device *device);

/**
 * Specify how many qp's are available for use
 *
 * @param [in] sha
 * The SHA engine instance to use.
 *
 * @param [in] num_qps
 * Number of QP's that can be used. Must be at least 1.
 *
 * @return
 * 0 on success or a negative status code on failure.
 *
 * @note this function shall only be called when the doca_sha instance is not running. see
 * doca_sha_start, doca_sha_stop
 *
 * @note this function is not thread safe. Application must ensure that this is only called from a
 * single thread.
 */
__DOCA_EXPERIMENTAL
int doca_sha_num_qps_set(struct doca_sha *sha, uint16_t num_qps);

/**
 * Enqueue jobs.
 *
 * @note
 * This function is thread safe when each queue is written to by only one
 * thread. If there is a chance that more than one thread could write to the
 * same queue then external synchronisation must be provided by the application
 * to prevent that circumstance.
 *
 * @param [in] sha
 * The SHA engine instance to use.
 *
 * @param [in] qid
 * ID of queue to enqueue the job into.
 *
 * @param [in] jobs
 * The jobs to enqueue.
 *
 * @param [in] jobs_count
 * The number of jobs to enqueue.
 *
 * @return
 * [0..jobs_count] Number of jobs enqueued or a negative posix status code.
 */
__DOCA_EXPERIMENTAL
int doca_sha_enqueue(struct doca_sha *sha, uint16_t qid,
			struct doca_sha_job_request *jobs, uint32_t jobs_count);

/**
 * Dequeue responses from the SHA engine.
 *
 * @note
 * This function is thread safe when each queue is read by only one thread. If
 * there is a chance that more than one thread could read from to the same queue
 * then external synchronisation must be provided by the application to prevent
 * that circumstance.
 *
 * @param [in] sha
 * The SHA engine to dequeue data from.
 *
 * @param [in] qid
 * ID of queue to read results from.
 *
 * @param [in,out] responses
 * A pointer to an array of responses. The SHA engine will place each dequeued result into
 * this array until either the maximum number of responses are returned or all available responses
 * have been returned. Caller assumes ownership of each response returned this way. The engine takes
 * no care to clear any un-populated element of this array, it is the caller responsibility to
 * ensure they only read as many responses as was indicated in the functions return value.
 *
 * @param [in] max_results
 * Maximum number of results to return. the responses array MUST have capacity
 * for at least this many elements.
 *
 * @return
 * [0..max_results] The number of responses dequeued or a negative posix status
 * code.
 */
__DOCA_EXPERIMENTAL
int doca_sha_dequeue(struct doca_sha *sha, uint16_t qid,
		       struct doca_sha_job_response *responses, uint16_t max_results);

/**
 * Get a new instance of a pre-configured implementation of a doca_sha_device.
 *
 * @details
 * DOCA sha provides some sha_impl implementations that are ready to use. This saves on the
 * effort of developing them yourself. You are free to disregard these and write your own to link
 * doca_sha to another sha implementation or if you need something custom to your needs.
 *
 * @return
 * A non NULL sha_impl on success, or NULL if the name is not recognised or some other error
 * occurred.
 */
__DOCA_EXPERIMENTAL
struct doca_sha_device *doca_sha_create_pre_configured_sha_impl(void);

/**
 * Convenience function to destroy a sha impl.
 *
 * @param [in] inst;
 * When this value is not NULL delegates the destruction to the impls destroy_fn
 *
 * @note
 * Will not call cleanup before destroying, it's assumed thats handled by the application.
 */
__DOCA_EXPERIMENTAL
void doca_sha_device_destroy(struct doca_sha_device *inst);

/** @} */

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DOCA_SHA_H_ */
