/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * Definition of an abstract implementation to be used by doca sha, Provides a separation between
 * implementation of sha routines for the higher level functionality of the doca sha api.
 */

#ifndef DOCA_SHA_DEVICE_H_
#define DOCA_SHA_DEVICE_H_

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Structure describing the capabilities supported by a sha impl.
 */
struct doca_sha_dev_caps {
	uint8_t multi_client_supported; /**< Device has multi-client support, 0 if not supported. */
};

/**
 * Structure describing a sha impl.
 */
struct doca_sha_dev_info {
	uint16_t version[3];   /**< version of the sha device [MAJOR][MINOR][PATCH]. */
	uint64_t max_job_size; /**< Maximum size of a job payload in bytes for a single job. */
	char const *name;      /**< A name describing this device. */
};

/**
 * Forward type declarations
 */
struct doca_sha_job_request;
struct doca_sha_job_response;

/**
 * Structure containing all the functionality required of a sha device by the DOCA sha.
 *
 * @details
 * Allows each DOCA sha instance to be supported by any sha_impl. Such as Hyperscan, DPDK sha,
 * or any other sha library that can be adapted to this interface.
 */
struct doca_sha_device {
	void *cookie;   /**< A place to bind any required extra data into the impl. */
	void *pd;       /**< Protection Domain assigned to device (only applicable to HW dev). */
	void *mem_regs; /**< Memory regions associated with the device (only applicable to HW dev). */

	/**
	 * @brief
	 * Do any one time setup work that may be required before performing sha operations.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] dev_addr
	 * PCIe address of sha device (NULL if not a HW device).
	 *
	 * @param [in] num_queues
	 * Number of qps to instantiate (0 if not a HW device).
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*init_fn)(struct doca_sha_device *inst, const char *dev_addr, const int num_queues);

	/**
	 * @brief
	 * Cleanup all resources acquired by the sha_impl from either one time setup or during
	 * runtime operation.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*cleanup_fn)(struct doca_sha_device *inst);

	/**
	 * @brief
	 * Function suitable to destroy this doca_sha_device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @details
	 * Set this to NULL to indicate that the engine should not attempt to destroy this object
	 * when the engine is destroyed.
	 */
	void (*destroy_fn)(struct doca_sha_device *inst);

	/**
	 * @brief
	 * Query capabilities of the sha device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [out] caps
	 * Capabilities structure to populate. MUST NOT BE NULL.
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*query_dev_caps_fn)(struct doca_sha_device *inst, struct doca_sha_dev_caps *caps);

	/**
	 * @brief
	 * Query information about the sha device.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [out] info
	 * Informational structure to populate. MUST NOT BE NULL.
	 *
	 * @return
	 * 0 on success or a negative posix status code.
	 */
	int (*query_dev_info_fn)(struct doca_sha_device *inst, struct doca_sha_dev_info *info);

	/**
	 * @brief
	 * Send a job to the sha engine to be processed.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in] qpid
	 * ID of qp to use
	 *
	 * @param [in] job
	 * Details of the job. Caller retains ownership of these objects;
	 *
	 * @param [in]
	 * Internal job id - this value should be returned by the device when dequeuing the result
	 * for this piece. The job request id is the id of the overall job which may have been
	 * broken into multiple pieces.
	 *
	 * @return
	 * [0..count] Number of jobs enqueued or a negative posix status code.
	 */
	int (*enqueue_job_fn)(struct doca_sha_device *inst, uint16_t qpid,
			      struct doca_sha_job_request **job, uint64_t *internal_job_ids, uint16_t jobs_count);

	/**
	 * @brief
	 * Fetch job responses from the sha_impl.
	 *
	 * @param [in] inst
	 * Instance of the impl.
	 *
	 * @param [in,out] responses
	 * A pointer to an array of responses. The sha engine will place each dequeued result into
	 * this array until either the maximum number of responses are returned or all available
	 * responses have been returned
	 *
	 * @param [in] max_responses
	 * The maximum number of responses the engine is able to accept from the sha_impl.
	 *
	 * @return
	 * [0..max_results] Number of jobs dequeued or a negative posix status code.
	 */
	int (*dequeue_jobs_fn)(struct doca_sha_device *inst, uint16_t qpid,
			       struct doca_sha_job_response *responses, uint16_t max_results);
};


#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DOCA_SHA_DEVICE_H_ */
